import logging
import os.path
import sys

from rook.processor.namespaces.python_object_namespace import PythonObjectNamespace
from rook.processor.namespaces.container_namespace import ContainerNamespace


class LogRecordNamespace(PythonObjectNamespace):

    def __init__(self, record):
        super(LogRecordNamespace, self).__init__(record, methods=self.METHODS)
        self.record = None

    def format(self, args=None):
        if self.record is None:
            self.obj = self.calc_record()
        return PythonObjectNamespace(self.obj.getMessage())
    
    def read_attribute(self, name):
        if self.record is None:
            self.obj = self.calc_record()
        return super(LogRecordNamespace, self).read_attribute(name)

    @staticmethod
    def find_caller():
        _logging_srcfile = logging._srcfile

        f = sys._getframe(8)
        while hasattr(f, "f_code") and f.f_code.co_name != "_callback":
            f = f.f_back

        # On some versions of IronPython, currentframe() returns None if
        # IronPython isn't run with -X:Frames.
        if f is not None:
            f = f.f_back

        rv = "(unknown file)", 0, "(unknown function)"
        while hasattr(f, "f_code"):
            co = f.f_code
            filename = os.path.normcase(co.co_filename)
            if filename == _logging_srcfile:
                f = f.f_back
                continue

            rv = co.co_filename, f.f_lineno, co.co_name
            break

        return rv

    def calc_record(self):
        logger, level, fn, lno, msg, args, exc_info, func, extra = self.obj

        if fn is None:
            try:
                fn, lno, func = self.find_caller()
            except ValueError:
                fn, lno, func = "(unknown file)", 0, "(unknown function)"

        if exc_info:
            if not isinstance(exc_info, tuple):
                exc_info = sys.exc_info()

        self.record = logger.makeRecord(logger.name, level, fn, lno, msg, args, exc_info, func, extra)
        return self.record

    def dump(self, args=None):
        if self.record is None:
            self.obj = self.calc_record()
        return ContainerNamespace({
            'name': PythonObjectNamespace(self.obj.name),
            'msg': PythonObjectNamespace(self.obj.msg),
            'formatted_message': self.format(),
            'args': PythonObjectNamespace(self.obj.args),
            'level_name': PythonObjectNamespace(self.obj.levelname),
            'level_no': PythonObjectNamespace(self.obj.levelno),
            'path_name': PythonObjectNamespace(self.obj.pathname),
            'filename': PythonObjectNamespace(self.obj.filename),
            'module': PythonObjectNamespace(self.obj.module),
            'lineno': PythonObjectNamespace(self.obj.lineno),
            'function': PythonObjectNamespace(self.obj.funcName),
            'time': PythonObjectNamespace(self.obj.created),
            'thread_id': PythonObjectNamespace(self.obj.thread),
            'thread_name': PythonObjectNamespace(self.obj.threadName),
            'process_name': PythonObjectNamespace(self.obj.processName),
            'process_id': PythonObjectNamespace(self.obj.process)
        })

    METHODS = (format, dump)



"""
Loads Rook into pyspark workers
Usage: spark-submit --conf spark.python.daemon.module=rook.pyspark_daemon
"""
import inspect
import os

import pyspark.daemon
import functools
import six
import sys

from rook.config import ImportServiceConfig

original_worker_main = pyspark.daemon.worker_main


def worker_main(*args, **kwargs):
    import rook
    try:
        rook.start(log_file="", log_to_stderr=True)
        from rook.logger import logger
        logger.debug("Started Rook in Spark worker")
        from rook.interface import _rook as singleton, _TRUE_VALUES
        from rook.services import ImportService
        import_service = singleton.get_trigger_services().get_service(ImportService.NAME)

        def try_load_rdd_udf_module(obj):
            try:
                # The pickled object that gets sent when you call a UDF on a RDD is a 4-elements tuple, with the first
                # one containing the function
                if type(obj) is tuple and len(obj) == 4 and callable(obj[0]):
                    # This looks complicated, but the UDF is stored as part of the closure, so we fetch it from the RDD
                    # object (there are several layers to traverse here...)
                    module_name = obj[0].__closure__[0].cell_contents.__closure__[0].cell_contents.__module__

                    # Sometimes the module's path isn't in sypath due to running pyspark in different working
                    # directory than the driver, so we add it just in case
                    path_to_module = os.path.dirname(inspect.getfile(obj[0].__closure__[0].cell_contents.__closure__[0].cell_contents))
                    sys.path.insert(0, path_to_module)
                    import_service.disable_evaluator()
                    import importlib
                    importlib.import_module(module_name)
                    return True
            except Exception as e:  # lgtm[py/catch-base-exception]
                logger.debug("pyspark udf (rdd): failed to read object {0} with error {1}".format(obj, str(e)))
            finally:
                import_service.enable_evaluator()

            return False

        def try_load_df_udf_module(obj):
            try:
                # The pickled object that gets sent when you call a UDF on a DF is a 2-elements tuple, with the first
                # one containing the function
                if type(obj) is tuple and len(obj) == 2 and callable(obj[0]):
                    # This looks complicated, but the UDF is stored as part of the closure, so we fetch it from the RDD
                    # object (there are several layers to traverse here...)
                    module_name = obj[0].__closure__[0].cell_contents.__module__

                    # Sometimes the module's path isn't in sypath due to running pyspark in different working
                    # directory than the driver, so we add it just in case
                    path_to_module = os.path.dirname(inspect.getfile(obj[0].__closure__[0].cell_contents))
                    sys.path.insert(0, path_to_module)
                    import_service.disable_evaluator()
                    import importlib
                    importlib.import_module(module_name)
                    return True
            except Exception as e:  # lgtm[py/catch-base-exception]
                logger.debug("pyspark udf (df): failed to read object {0} with error {1}".format(obj, str(e)))
            finally:
                import_service.enable_evaluator()

            return False

        def pickle_load_hook(orig_func, eager_load, *args, **kwargs):
            obj = orig_func(*args, **kwargs)
            loaded_udf = try_load_rdd_udf_module(obj)
            if not loaded_udf:
                loaded_udf = try_load_df_udf_module(obj)
            if eager_load or loaded_udf:
                try:
                    # this is here to deal with the delay of having the periodic thread call evaluate_module_list -
                    # it could miss a module being imported.
                    import_service.evaluate_module_list()
                except:  # lgtm[py/catch-base-exception]
                    logger.exception("Silenced exception during module list evaluation")
            return obj

        # we may end up missing pickle module imports if we rely on the sys.modules polling thread
        # only do this eagerly if we're not using the import hook
        should_eager_load = ImportServiceConfig.USE_IMPORT_HOOK is False
        import pyspark.serializers
        pyspark.serializers.pickle.loads = functools.partial(pickle_load_hook, pyspark.serializers.pickle.loads,
                                                             should_eager_load)
        pyspark.serializers.pickle.load = functools.partial(pickle_load_hook, pyspark.serializers.pickle.load,
                                                            should_eager_load)
    except:  # lgtm[py/catch-base-exception]
        six.print_("Starting Rook in worker_main failed", file=sys.stderr)

    result = original_worker_main(*args, **kwargs)
    try:
        rook.flush()
    except:  # lgtm[py/catch-base-exception]
        pass
    return result


pyspark.daemon.worker_main = worker_main

if __name__ == '__main__':
    pyspark.daemon.manager()



import collections

LIST_TYPE = [list, tuple, set, frozenset]
DICT_TYPE = [dict, ]
try:
    import numpy
    LIST_TYPE.append(numpy.ndarray)
except ImportError:
    numpy = None

LIST_NAMES = ('deque', 'UserList')

DICT_NAMES = ('ChainMap', 'Counter', 'OrderedDict', 'defaultdict', 'UserDict')

for list_name in LIST_NAMES:
    try:
        LIST_TYPE.append(getattr(collections, list_name))
    except AttributeError:
        pass  # lgtm - ignore missing type

for dict_name in DICT_NAMES:
    try:
        DICT_TYPE.append(getattr(collections, dict_name))
    except AttributeError:
        pass  # lgtm - ignore missing type

LIST_TYPE = tuple(LIST_TYPE)
DICT_TYPE = tuple(DICT_TYPE)



import datetime
import decimal
from types import FunctionType, MethodType, ModuleType
from google.protobuf.reflection import GeneratedProtocolMessageType
from google.protobuf.internal import api_implementation

if api_implementation.Type() == "upb":
    import google

import six

if six.PY2:
    from types import TypeType

try:
    import numpy
except ImportError:
    numpy = None

try:
    import torch
except ImportError:
    torch = None

try:
    import multidict
except ImportError:
    multidict = None

try:
    from bson import ObjectId
    from bson.raw_bson import RawBSONDocument
except ImportError:
    ObjectId = None
    RawBSONDocument = None


class NamespaceSerializerBase(object):
    BUILTIN_ATTRIBUTES_IGNORE = {
        '__dict__',
        '__module__',
        '__weakref__',
        '__name__',
        '__doc__',
        '__qualname__',
        '__spec__',
        '__defaults__',
        '__code__',
        '__globals__',
        '__closure__',
        '__annotations__',
        '__kwdefaults__',
        '__bases__'}

    GETATTR_ALLOWLIST = {
        'py4j.java_gateway.JVMView'
    }

    try:
        # Py2 objects
        BINARY_TYPES = (buffer, bytearray)
        STRING_TYPES = (basestring, str, unicode)
        CODE_TYPES = (FunctionType, MethodType, TypeType, ModuleType, six.MovedModule)
        PRIMITIVE_TYPES = (type(None), int, bool, long, float, str, unicode, complex, decimal.Decimal) + BINARY_TYPES + CODE_TYPES + (datetime.datetime,)

    except NameError:
        # Py3 objects
        BINARY_TYPES = (bytearray, bytes, bytes)
        STRING_TYPES = (str, )
        CODE_TYPES = (FunctionType, MethodType, type, ModuleType, six.MovedModule)
        PRIMITIVE_TYPES = (type(None), int, bool, float, str, complex, decimal.Decimal) + BINARY_TYPES + CODE_TYPES + (datetime.datetime,)

    def __init__(self, use_string_cache=False):
        self.use_string_cache = use_string_cache
        self.string_cache = {}

        if use_string_cache:
            # Lock the 0 index since some variant will have no originalType (container for example)
            self.string_cache[""] = 0

        self.estimated_pending_bytes = 0

    def get_string_cache(self):
        return self.string_cache

    def get_estimated_pending_bytes(self):
        return self.estimated_pending_bytes

    def _get_string_index_in_cache(self, string_to_cache):
        index = self.string_cache.get(string_to_cache, None)
        if index is None:
            index = len(self.string_cache)
            # We estimate each character is one byte in utf-8 and overhead is 5 bytes
            self.estimated_pending_bytes += len(string_to_cache) + 5
            self.string_cache[string_to_cache] = index

        return index

    def is_in_get_attr_allowlist(self, obj_name):
        return obj_name in self.GETATTR_ALLOWLIST

    if six.PY3:
        @staticmethod
        def get_attr_safe(obj, name):
            try:
                return object.__getattribute__(obj, name)
            except AttributeError:
                return None

        @staticmethod
        def has_attr_safe(obj, name):
            try:
                object.__getattribute__(obj, name)
                return True
            except AttributeError:
                return False
    else:
        # In python 2, you don't technically have to inherit from `object`, so some methods aren't using
        # `__getattribute__` like object does. Therefore, we must use the unsafe alternatives
        class OldStyleClass:
            pass
        instance = type(OldStyleClass())

        @staticmethod
        def get_attr_safe(obj, name):
            try:
                return object.__getattribute__(obj, name)
            except AttributeError:
                try:
                    # If obj inherits from object, we are done
                    if type(obj) is not NamespaceSerializerBase.instance:
                        return None
                    # Else, use the unsafe method
                    return getattr(obj, name)
                # If an exception is thrown from `__getattr__`, the type of the exception will not
                # be `AttributeError`, so we must catch 'em all
                except Exception as e:
                    return None

        @staticmethod
        def has_attr_safe(obj, name):
            return hasattr(obj, name)

    def _get_object_width(self, obj):
        object_width = 0
        obj_dict = self.get_attr_safe(obj, '__dict__')
        obj_slots = self.get_attr_safe(obj, '__slots__')

        if obj_dict:
            object_width += len(obj_dict)
        if obj_slots:
            object_width += len(obj_slots)
        return object_width

    if six.PY2:
        @staticmethod
        def normalize_string(obj):
            if isinstance(obj, str):
                return unicode(obj, errors="replace")
            else:
                return unicode(obj)
    else:
        @staticmethod
        def normalize_string(obj):
            return obj

    if numpy is not None:
        @staticmethod
        def is_numpy_obj(obj_mro):
            return numpy.generic in obj_mro

        @staticmethod
        def is_numpy_array(obj_class):
            return obj_class is numpy.ndarray

    else:
        @staticmethod
        def is_numpy_obj(obj_mro):
            return False

        @staticmethod
        def is_numpy_array(obj_class):
            return False

    if torch is not None:
        @staticmethod
        def is_torch_obj(obj_class):
            try:
                module = object.__getattribute__(obj_class, '__module__')
            except AttributeError:
                return False

            return module.startswith('torch')
    else:
        @staticmethod
        def is_torch_obj(obj_class):
            return False

    if multidict is not None:
        @staticmethod
        def is_multidict_obj(obj_class):
            return obj_class in (multidict.MultiDict, multidict.CIMultiDict,
                                 multidict.MultiDictProxy,
                                 multidict.CIMultiDictProxy)
    else:
        @staticmethod
        def is_multidict_obj(obj_class):
            return False

    if api_implementation.Type() == "upb":
        @staticmethod
        def is_protobuf_obj(obj_class):
            return isinstance(obj_class, google._upb._message.MessageMeta)
    else:
        @staticmethod
        def is_protobuf_obj(obj_class):
            return isinstance(obj_class, GeneratedProtocolMessageType)

    if not ObjectId or not RawBSONDocument:
        @staticmethod
        def is_bson(obj_class):
            return False
    else:
        @staticmethod
        def is_bson(obj_class):
            return obj_class in (ObjectId, RawBSONDocument)
